def pretty_table(headers, rows):
    """
    :param headers: A list, the column names.
    :param rows: A list of lists, the row data.
    """

    rows = [[stringify(s) for s in row] for row in rows]

    if not all([len(headers) == len(row) for row in rows]):
        return "Incorrect number of rows."

    columns = [[headers[i]] + [row[i] for row in rows] for i in range(len(headers))]
    widths = [len(max(l, key=len)) for l in columns]

    top = [x.ljust(widths[i]) for i, x in enumerate(headers)]
    separator = ["-" * widths[i] for i in range(len(headers))]
    rest = [[x.ljust(widths[i]) for i, x in enumerate(item)] for item in rows]

    top = " | ".join(top)
    separator = "-+-".join(separator)
    rest = [" | ".join(r) for r in rest]

    top += "\n"
    separator += "\n"
    rest = "\n".join(rest)

    return top + separator + rest + "\n"

def stringify(s):
    if s is None:
        return ""

    try:
        unicode
    except NameError:
        return str(s)
    else:
        if isinstance(s, unicode):
            return s.encode("utf8")

    return str(s)



import re
from pygments.lexer import RegexLexer
from pygments.token import Text, Comment, Operator, Keyword, Name, String, Number, Token

__all__ = ["CypherLexer"]

class CypherLexer(RegexLexer):
    name = 'Cypher'
    aliases = ['cypher']
    filenames = ['*.cyp', '*.cypher']

    flags = re.IGNORECASE
    tokens = {
        'root': [
            (r'\s+', Text),
            (r'//.*?\n', Comment.Single),
            (r'\b(ABS|ACOS|ALL|ALLSHORTESTPATHS|AND|ANY|AS|ASC|ASCENDING|ASIN|ASSERT|ATAN|ATAN2|AVG'
             r'|BY'
             r'|CASE|CEIL|COALESCE|COLLECT|COMMIT|CONSTRAINT|CONTAINS|COS|COT|COUNT|CREATE|CYPHER'
             r'|DEGREES|DELETE|DESC|DESCENDING|DETACH|DISTINCT|DROP'
             r'|E|ELSE|END|ENDNODE|ENDS|EXISTS|EXP|EXPLAIN|EXTRACT'
             r'|FALSE|FIELDTERMINATOR|FILTER|FLOOR|FOREACH|FROM'
             r'|HAS|HAVERSIN|HEAD|HEADERS'
             r'|ID|ILIKE|IN|INDEX|IS'
             r'|KEYS'
             r'|LABELS|LAST|LEFT|LENGTH|LIKE|LIMIT|LOAD CSV|LOG|LOG10|LOWER|LTRIM'
             r'|MATCH|MAX|MERGE|MIN'
             r'|NODE|NODES|NONE|NOT|NULL'
             r'|ON|OPTIONAL|OR|ORDER'
             r'|PERCENTILECONT|PERCENTILEDISC|PERIODIC|PI|PROFILE'
             r'|RADIANS|RAND|RANGE|REDUCE|REL|RELATIONSHIP|RELATIONSHIPS|REMOVE|REPLACE|RETURN|RIGHT|ROUND|RTRIM'
             r'|SCAN|SET|SHORTESTPATH|SIGN|SIN|SINGLE|SIZE|SKIP|SPLIT|SQRT|START|STARTNODE|STARTS|STDEV|STDEVP|STR|SUBSTRING|SUM'
             r'|TAIL|TAN|THEN|TIMESTAMP|TOFLOAT|TOINT|TRIM|TRUE|TYPE'
             r'|UNION|UNIQUE|UNWIND|USING|UPPER'
             r'|WHEN|WHERE|WITH'
             r'|XOR)\b', Keyword),
            (r'[+*/<>=~!@#%^&|`?-]', Operator),
            (r'[0-9]+', Name),
            (r"'(''|[^'])*'", String.Single),
            (r'"(""|[^"])*"', String.Symbol),
            (r'[a-zA-Z_][a-zA-Z0-9_]*', Name),
            (r'[\]\)]-[\(\[]|[\]\)]->\(|\)<-[\[\(]|[\)\]]-->\(|\)<--[\(\[]|[\)\]]--[\(\[]', Token.Pattern),
            (r'\.', Token.Pattern),
            (r'\(|\)|\]|\[|{|}', Token.Pattern)
        ]
    }


from __future__ import unicode_literals, print_function

import sys
import re
import os
import csv
from datetime import datetime

import click
from prompt_toolkit import Application, CommandLineInterface, AbortAction
from prompt_toolkit.buffer import AcceptAction
from prompt_toolkit.history import FileHistory
from prompt_toolkit.shortcuts import create_prompt_layout, create_eventloop
from prompt_toolkit.styles import PygmentsStyle
from pygments.token import Token

from cycli import __version__
from cycli.lexer import CypherLexer
from cycli.style import CypherStyle
from cycli.completer import CypherCompleter
from cycli.buffer import CypherBuffer
from cycli.binder import CypherBinder
from cycli.driver import Neo4j, AuthError, ConnectionError
from cycli.table import pretty_table
from cycli.cypher import Cypher


def get_tokens(x):
        return [(Token.Prompt, "> ")]


class Cycli:

    def __init__(self, host, port, username, password, logfile, filename, ssl, read_only, timeout):
        self.logfile = logfile
        self.filename = filename
        self.read_only = read_only
        self.neo4j = Neo4j(host, port, username, password, ssl, timeout)
        self.cypher = Cypher()

    def write_to_logfile(self, query, response):
        headers = response["headers"]
        rows = response["rows"]
        duration = response["duration"]
        error = response["error"]

        self.logfile.write("> {}\n".format(query))
        self.logfile.write("{}\n".format(pretty_table(headers, rows)))

        if error is False:
            self.logfile.write("{} ms\n\n".format(duration))

    @staticmethod
    def write_to_csvfile(headers, rows):
        filename = "cycli {}.csv".format(datetime.now().strftime("%Y-%m-%d at %I.%M.%S %p"))

        with open(filename, "wt") as csvfile:
            csvwriter = csv.writer(csvfile, quotechar=str('"'), quoting=csv.QUOTE_NONNUMERIC, delimiter=str(","))
            csvwriter.writerow(headers)

            for row in rows:
                csvwriter.writerow(row)

        csvfile.close()

    def run(self):
        labels = self.neo4j.get_labels()
        relationship_types = self.neo4j.get_relationship_types()
        properties = self.neo4j.get_property_keys()

        if self.filename:
            queries = self.filename.read()
            queries = queries.split(";")[:-1]

            for query in queries:
                query += ";"
                query = query.strip()

                print("> " + query)
                self.handle_query(query)
                print()

            return

        click.secho(" ______     __  __     ______     __         __    ", fg="red")
        click.secho("/\  ___\   /\ \_\ \   /\  ___\   /\ \       /\ \   ", fg="yellow")
        click.secho("\ \ \____  \ \____ \  \ \ \____  \ \ \____  \ \ \  ", fg="green")
        click.secho(" \ \_____\  \/\_____\  \ \_____\  \ \_____\  \ \_\ ", fg="blue")
        click.secho("  \/_____/   \/_____/   \/_____/   \/_____/   \/_/ ", fg="magenta")

        msg = "\nUsing Bolt." if type(self.neo4j.client).__name__ == "BoltClient" else ""

        print(msg)
        print("Version: {}".format(__version__))
        print("Bug reports: https://github.com/nicolewhite/cycli/issues\n")

        completer = CypherCompleter(labels, relationship_types, properties)

        layout = create_prompt_layout(
            lexer=CypherLexer,
            get_prompt_tokens=get_tokens,
            reserve_space_for_menu=8,
        )

        buff = CypherBuffer(
            accept_action=AcceptAction.RETURN_DOCUMENT,
            history=FileHistory(filename=os.path.expanduser('~/.cycli_history')),
            completer=completer,
            complete_while_typing=True,
        )

        application = Application(
            style=PygmentsStyle(CypherStyle),
            buffer=buff,
            layout=layout,
            on_exit=AbortAction.RAISE_EXCEPTION,
            key_bindings_registry=CypherBinder.registry
        )

        cli = CommandLineInterface(application=application, eventloop=create_eventloop())

        try:
            while True:
                document = cli.run()
                query = document.text
                self.handle_query(query)

        except:
            print("Goodbye!")

    def handle_query(self, query):
        run_n = re.match('run-([0-9]+) (.*)', query, re.DOTALL)
        save_csv = query.startswith("save-csv ")

        if self.cypher.is_a_write_query(query) and self.read_only:
            print("Query aborted. You are in read-only mode.")

        elif query in ["quit", "exit"]:
            raise Exception

        elif query == "help":
            print_help()

        elif query == "refresh":
            self.neo4j.refresh()

        elif query == "schema":
            self.neo4j.print_schema()

        elif query == "schema-indexes":
            self.neo4j.print_indexes()

        elif query == "schema-constraints":
            self.neo4j.print_constraints()

        elif query == "schema-labels":
            self.neo4j.print_labels()

        elif query == "schema-rels":
            self.neo4j.print_relationship_types()

        elif query.startswith("env"):
            if query == "env":
                for key, value in self.neo4j.parameters.items():
                    print("{0}={1}".format(key, value))
            else:
                key = query[3:]
                key = key.strip("'\"[]")
                value = self.neo4j.parameters.get(key)

                if value is not None:
                    print(value)

        elif query.startswith("export "):
            if "=" not in query:
                print("Set parameters with export key=value.")
            else:
                params = query.replace("export ", "").strip()
                key, value = params.split("=", 1)
                key = key.strip()
                value = value.strip()

                try:
                    value = eval(value)
                    self.neo4j.update_parameters(key, value)
                except Exception as e:
                    print(e)

        else:
            count = int(run_n.group(1)) if run_n else 1
            query = run_n.group(2) if run_n else query
            query = query[len("save-csv "):] if save_csv else query

            if count <= 0 or not query:
                raise Exception

            total_duration = 0
            index = 0

            while index < count:
                response = self.neo4j.cypher(query)

                headers = response["headers"]
                rows = response["rows"]
                duration = response["duration"]
                error = response["error"]

                if error is False:
                    print(pretty_table(headers, rows))

                    ms = "Run {}: {} ms\n".format(index + 1, duration) if run_n else "{} ms".format(duration)
                    print(ms)

                    if save_csv:
                        self.write_to_csvfile(headers, rows)
                else:
                    print(error)

                if self.logfile:
                    self.write_to_logfile(query, response)

                total_duration += duration
                index += 1

            if run_n and error is False:
                print("Total duration: {} ms".format(total_duration))


def print_help():
    headers = ["Keyword", "Description"]

    rows = [
        ["quit", "Exit cycli."],
        ["exit", "Exit cycli."],
        ["help", "Display this text."],
        ["refresh", "Refresh schema cache."],
        ["run-n", "Run a Cypher query n times."],
        ["export", "Set a parameter with export key=value."],
        ["save-csv", "Save the query results to a CSV file."],
        ["schema", "Display indexes, constraints, labels, and relationship types."],
        ["schema-indexes", "Display indexes."],
        ["schema-constraints", "Display constraints."],
        ["schema-labels", "Display labels."],
        ["schema-rels", "Display relationship types."],
        ["CTRL-D", "Exit cycli if the input is blank."],
        ["CTRL-C", "Abort and rollback the currently-running query."]
    ]

    print(pretty_table(headers, rows))


@click.command()
@click.option("-v", "--version", is_flag=True, help="Show cycli version and exit.")
@click.option("-h", "--host", default="localhost", help="The host address of Neo4j.")
@click.option("-P", "--port", default="7474", help="The port number on which Neo4j is listening.")
@click.option("-u", "--username", help="Username for Neo4j authentication.")
@click.option("-p", "--password", help="Password for Neo4j authentication.")
@click.option("-t", "--timeout", help="Set a global socket timeout for queries.", type=click.INT)
@click.option('-l', '--logfile', type=click.File(mode="a", encoding="utf-8"), help="Log every query and its results to a file.")
@click.option("-f", "--filename", type=click.File(mode="rb"), help="Execute semicolon-separated Cypher queries from a file.")
@click.option("-s", "--ssl", is_flag=True, help="Use the HTTPS protocol.")
@click.option("-r", "--read-only", is_flag=True, help="Do not allow any write queries.")
def run(host, port, username, version, timeout, password, logfile, filename, ssl, read_only):
    if version:
        print("cycli {}".format(__version__))
        sys.exit(0)

    if username and not password:
        password = click.prompt("Password", hide_input=True, show_default=False, type=str)

    try:
        cycli = Cycli(host, port, username, password, logfile, filename, ssl, read_only, timeout)

    except AuthError:
        print("Unauthorized. See cycli --help for authorization instructions.")

    except ConnectionError:
        print("Connection refused. Is Neo4j turned on?")

    else:
        cycli.run()


if __name__ == '__main__':
    run()



from __future__ import unicode_literals
from prompt_toolkit.completion import Completer, Completion
from cycli.cypher import Cypher


class CypherCompleter(Completer):

    def __init__(self, labels, relationship_types, properties):
        super(CypherCompleter, self).__init__()

        self.labels = labels
        self.relationship_types = relationship_types
        self.properties = properties

        self.cypher = Cypher()

    def get_completions(self, document, complete_event):
        text_before_cursor = document.text_before_cursor
        choices = []
        lookup = ""

        if self.exists_unclosed_char("'", text_before_cursor) or self.exists_unclosed_char('"', text_before_cursor):
            return
        elif self.typing_label(text_before_cursor):
            choices = self.labels
            lookup = self.everything_after_last(":", text_before_cursor)
        elif self.typing_relationship(text_before_cursor):
            choices = self.relationship_types
            lookup = self.everything_after_last(":", text_before_cursor)
        elif self.typing_property(text_before_cursor):
            loc = text_before_cursor.rfind(".")
            identifier = text_before_cursor[text_before_cursor.rfind(" ", 0, loc) + 1:loc]

            if identifier.isalnum() and any(c.isalpha() for c in identifier):
                choices = self.properties
                lookup = self.everything_after_last(".", text_before_cursor)
            else:
                return
        elif text_before_cursor:
            if text_before_cursor[-1].isalpha():
                last_cypher_word = self.most_recent_cypher_word(text_before_cursor)
                choices = self.cypher.most_probable_next_keyword(last_cypher_word)
                lookup = self.last_alphabetic_chunk(text_before_cursor)
        else:
            return

        completions = self.find_matches(lookup, choices)

        for completion in completions:
            yield Completion(completion, -len(lookup))

    @staticmethod
    def find_matches(word, choices):
        word = word.lower()
        lower_choices = [x.lower() for x in choices]

        completions = []

        for i, choice in enumerate(lower_choices):
            if choice.startswith(word):
                completions.append(choices[i])

        return completions

    def most_recent_cypher_word(self, chars):
        text = " " + chars
        keyword_indices = [(word, text.rfind(" " + word + " ")) for word in self.cypher.KEYWORDS]
        function_indices = [(word, text.rfind(" " + word + "(")) for word in self.cypher.FUNCTIONS]

        indices = keyword_indices + function_indices

        # If no keywords were found, we want to be in the "" state in the Markov model.
        if not any([i[1] > -1 for i in indices]):
            return ""

        most_recent = max(indices, key=lambda i:i[1])[0]
        return most_recent

    @staticmethod
    def last_alphabetic_chunk(chars):
        chars = list(chars)
        chars.reverse()

        keep = []

        for c in chars:
            if c.isalpha():
                keep.append(c)
            else:
                break

        keep.reverse()
        return "".join(keep)

    @staticmethod
    def everything_after_last(char, chars):
        chars = chars.replace("`", "")
        loc = chars.rfind(char)
        return chars[loc + 1:]

    @staticmethod
    def exists_unclosed_char(char, chars):
        return chars.count(char) % 2 != 0

    @staticmethod
    def exists_unclosed_pattern(open_char, close_char, chars):
        return chars.count(open_char) != chars.count(close_char)

    def colon_inside_unclosed_pattern(self, open_char, close_char, chars):
        return self.exists_unclosed_pattern(open_char, close_char, chars) and chars.rfind(":") > chars.rfind(open_char)

    def typing_relationship(self, chars):
        return self.colon_inside_unclosed_pattern("[", "]", chars)

    def typing_label(self, chars):
        return self.colon_inside_unclosed_pattern("(", ")", chars)

    def typing_property(self, chars):
        chars = list(chars)
        chars.reverse()

        skip = ["_", "`"]

        # Skip spaces if we're inside backticks.
        if self.exists_unclosed_char("`", chars):
            skip.append(" ")

        for c in chars:
            if not c.isalnum() and c not in skip:
                return c == "."

        return False




from datetime import datetime
from cycli.table import pretty_table


class AuthError(Exception):
    pass


class ConnectionError(Exception):
    pass


class Neo4j:

    Client = None

    relationship_types = None
    labels = None
    constraints = None
    indexes = None

    parameters = {}

    def __init__(self, host, port, username=None, password=None, ssl=False, timeout=None):
        self.client = self.Client(host, port, username, password, ssl, timeout)

    def cypher(self, query):
        return self.client.cypher(query, self.parameters)

    def get_labels(self):
        if not self.labels:
            self.labels = self.client.get_labels()
        return self.labels

    def get_relationship_types(self):
        if not self.relationship_types:
            self.relationship_types = self.client.get_relationship_types()
        return self.relationship_types

    def get_constraints(self):
        if not self.constraints:
            self.constraints = self.client.get_constraints()
        return self.constraints

    def get_indexes(self):
        if not self.indexes:
            self.indexes = self.client.get_indexes()
        return self.indexes

    def get_property_keys(self):
        return self.client.get_property_keys()

    def update_parameters(self, key, value):
        self.parameters[key] = value

    def refresh(self):
        self.labels = None
        self.relationship_types = None
        self.indexes = None
        self.constraints = None
        self.get_labels()
        self.get_relationship_types()
        self.get_indexes()
        self.get_constraints()

    def print_labels(self):
        headers = ["Labels"]
        rows = [[x] for x in self.get_labels()]

        print(pretty_table(headers, rows))

    def print_relationship_types(self):
        headers = ["Relationship Types"]
        rows = [[x] for x in self.get_relationship_types()]

        print(pretty_table(headers, rows))

    def print_constraints(self):
        headers = ["Constraints"]
        constraints = self.get_constraints()
        rows = [[x] for x in self.format_constraints_indexes(constraints)]

        print(pretty_table(headers, rows))

    def print_indexes(self):
        headers = ["Indexes"]
        indexes = self.get_indexes()
        rows = [[x] for x in self.format_constraints_indexes(indexes)]

        print(pretty_table(headers, rows))

    def format_constraints_indexes(self, values):
        return [":{}({})".format(value["label"], ",".join(value["property_keys"])) for value in values]

    def print_schema(self):
        headers = ["Labels", "Relationship Types", "Constraints", "Indexes"]

        columns = [self.get_labels()[:]]
        columns.append(self.get_relationship_types()[:])
        columns.append(self.format_constraints_indexes(self.get_constraints()[:]))
        columns.append(self.format_constraints_indexes(self.get_indexes()[:]))

        max_length = len(max(columns, key=len))
        [x.extend([""] * (max_length - len(x))) for x in columns]
        rows = [[x[i] for x in columns] for i in range(max_length)]

        print(pretty_table(headers, rows))


try:
    from neo4j.v1 import GraphDatabase
    from json import loads
    try:
        from urllib.request import urlopen
    except ImportError:
        from urllib import urlopen

    class BoltClient(object):

        def __init__(self, host, port, username=None, password=None, ssl=False, timeout=None):
            bolt_uri = "bolt://{host}".format(host=host, port=port)

            if username and password:
                # todo: set auth header
                pass

            self.http_uri = "http://{host}:{port}/db/data/".format(host=host, port=port)
            driver = GraphDatabase.driver(bolt_uri)
            self.session = driver.session()

        def cypher(self, statement, parameters):
            error = False
            headers = []
            rows = []

            start = datetime.now()
            tx = self.session.begin_transaction()

            try:
                cursor = tx.run(statement, parameters)
                rows = [list(row.values()) for row in cursor.stream()]
                headers = cursor.keys()
                tx.commit()
            except KeyboardInterrupt:
                tx.rollback()
                error = ""
            except Exception as e:
                error = e

            end = datetime.now()
            duration = int(round((end - start).total_seconds() * 1000))

            return {
                "headers": headers,
                "rows": rows,
                "duration": duration,
                "error": error
            }

        def get_labels(self):
            return sorted(loads(urlopen(self.http_uri + "labels").read()))

        def get_relationship_types(self):
            return sorted(loads(urlopen(self.http_uri + "relationship/types").read()))

        def get_constraints(self):
            return sorted(loads(urlopen(self.http_uri + "schema/constraint").read()))

        def get_indexes(self):
            return sorted(loads(urlopen(self.http_uri + "schema/index").read()))

        def get_property_keys(self):
            return sorted(loads(urlopen(self.http_uri + "propertykeys").read()))

    Neo4j.Client = BoltClient

except ImportError:
    from py2neo import Graph, authenticate, Unauthorized
    from py2neo.packages.httpstream import SocketError, http

    class Py2neoClient(object):

        def __init__(self, host, port, username=None, password=None, ssl=False, timeout=None):
            if timeout is not None:
                http.socket_timeout = timeout

            host_port = "{host}:{port}".format(host=host, port=port)
            uri = "{scheme}://{host_port}/db/data/".format(scheme="https" if ssl else "http", host_port=host_port)

            if username and password:
                authenticate(host_port, username, password)

            self.graph = Graph(uri)

            try:
                self.neo4j_version = self.graph.neo4j_version
            except Unauthorized:
                raise AuthError(uri)
            except SocketError:
                raise ConnectionError(uri)

        def cypher(self, statement, parameters):
            error = False
            headers = []
            rows = []

            start = datetime.now()
            tx = self.graph.cypher.begin()

            try:
                tx.append(statement, parameters)
                results = tx.process()
                results = results[0]
                headers = results.columns
                rows = [[x[header] for header in headers] for x in results]
                tx.commit()
            except KeyboardInterrupt:
                tx.rollback()
                error = ""
            except Exception as e:
                error = e

            end = datetime.now()
            duration = int(round((end - start).total_seconds() * 1000))

            return {
                "headers": headers,
                "rows": rows,
                "duration": duration,
                "error": error
            }

        def get_labels(self):
            return sorted(self.graph.resource.resolve("labels").get().content)

        def get_relationship_types(self):
            return sorted(self.graph.resource.resolve("relationship/types").get().content)

        def get_constraints(self):
            return sorted(self.graph.resource.resolve("schema/constraint").get().content)

        def get_indexes(self):
            return sorted(self.graph.resource.resolve("schema/index").get().content)

        def get_property_keys(self):
            return sorted(self.graph.resource.resolve("propertykeys").get().content)

    Neo4j.Client = Py2neoClient


from setuptools import setup
from cycli import __version__

setup(name='cycli',
      version=__version__,
      description='A Command Line Interface for Cypher.',
      long_description='Syntax highlighting and autocomplete.',
      keywords='neo4j cypher cli syntax autocomplete',
      url='https://github.com/nicolewhite/cycli',
      author='Nicole White',
      author_email='nicole@neo4j.com',
      license='MIT',
      packages=['cycli'],
      install_requires=[
        'click==4.1',
        'prompt-toolkit==0.57',
        'Pygments==2.0.2',
        'py2neo>=2.0,<3.0',
      ],
      include_package_data=True,
      zip_safe=False,
      entry_points={
        'console_scripts': [
            'cycli = cycli.main:run'
        ]
    })


from __future__ import unicode_literals
from prompt_toolkit.key_binding.manager import KeyBindingManager

CypherBinder = KeyBindingManager()

@CypherBinder.registry.add_binding("{")
def curly_left(event):
    b = event.cli.current_buffer
    b.insert_text("{")
    b.insert_text("}", move_cursor=False)

@CypherBinder.registry.add_binding("}")
def curly_right(event):
    b = event.cli.current_buffer
    char = b.document.current_char

    if char == "}":
        b.cursor_right()
    else:
        b.insert_text("}")

@CypherBinder.registry.add_binding("(")
def paren_left(event):
    b = event.cli.current_buffer
    b.insert_text("(")
    b.insert_text(")", move_cursor=False)

@CypherBinder.registry.add_binding(")")
def paren_right(event):
    b = event.cli.current_buffer
    char = b.document.current_char

    if char == ")":
        b.cursor_right()
    else:
        b.insert_text(")")

@CypherBinder.registry.add_binding("[")
def bracket_left(event):
    b = event.cli.current_buffer
    b.insert_text("[")
    b.insert_text("]", move_cursor=False)

@CypherBinder.registry.add_binding("]")
def bracket_right(event):
    b = event.cli.current_buffer
    char = b.document.current_char

    if char == "]":
        b.cursor_right()
    else:
        b.insert_text("]")

@CypherBinder.registry.add_binding("'")
def apostrophe(event):
    b = event.cli.current_buffer
    char = b.document.current_char

    if char == "'":
        b.cursor_right()
    else:
        b.insert_text("'")
        b.insert_text("'", move_cursor=False)

@CypherBinder.registry.add_binding("\"")
def quote(event):
    b = event.cli.current_buffer
    char = b.document.current_char

    if char == "\"":
        b.cursor_right()
    else:
        b.insert_text("\"")
        b.insert_text("\"", move_cursor=False)

@CypherBinder.registry.add_binding("`")
def backtick(event):
    b = event.cli.current_buffer
    char = b.document.current_char

    if char == "\"":
        b.cursor_right()
    else:
        b.insert_text("`")
        b.insert_text("`", move_cursor=False)